unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, Windows, SQLDB, ODBCConn, FileUtil;

type
  TForm1 = class; // forward-Deklaration, damit der Name schon für TMyThread bekannt ist.

  { TMyBlockingThread }
  TMyBlockingThread = class(TThread)
  private
  protected
    procedure Execute; override;
  public
    Query:      TSQLQuery;
    Constructor Create(CreateSuspended: boolean);
  end;

  { TMyThread }
  TMyThread = class(TThread)
  private
    fStatusText: String;
    fStartTime:  TDateTime;
    fForm:       TForm1;
    fRowCnt:     Integer;
    fSQLText:    TStringList;
    procedure ShowStatus;
    procedure FetchData;
    procedure ButtonStopEnable;
  protected
    procedure Execute; override;
  public
    Connection:  TODBCConnection;
    Query:       TSQLQuery;
    Transaction: TSQLTransaction;
    Constructor Create(CreateSuspended: boolean);
  end;

  { TForm1 }
  TForm1 = class(TForm)
    Button1: TButton;
    ComboBox1: TComboBox;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure OnThreadDone(Sender: TObject);
  private
  public
  end;

var
  Form1:            TForm1;
  MyThread:         TMyThread;
  MyBlockingThread: TMyBlockingThread;

implementation

{$R *.lfm}

function GetThreadIsRunning(aThread: TThread): Boolean;
begin
  if Assigned(aThread) then // prüfen, ob der Thread bereits erzeugt wurde
    Result := not aThread.Finished // Thread läuft noch oder wartet auf die Ausführung von OnThreadDone
  else
    Result := False; // Thread wurde noch nicht gestartet oder ist bereits fertig
end;

{ TForm1 }
procedure TForm1.Button1Click(Sender: TObject);
begin
  Screen.Cursor := crHourGlass;
  try
    Button1.Enabled := false;

    {if GetThreadIsRunning(MyBlockingThread) then
    begin
      //MyBlockingThread.Free;
      try
        //MyBlockingThread.Terminate;
        MyBlockingThread.Free;
        //CloseHandle(MyBlockingThread.Handle);
      except
        ;
      end;
      ShowMessage('MyBlockingThread is still running!');
    end;}

    if (not GetThreadIsRunning(MyThread)) then
    begin
      MyThread := TMyThread.Create(True); // This way it doesn't start automatically
      MyThread.OnTerminate := @OnThreadDone;
      MyThread.fForm := Self;
      MyThread.fSQLText.LoadFromFile(ComboBox1.Text);
      MyThread.Start;
      Button1.Caption := 'Please wait ...';
      //ShowMessage('Abfrage gestartet');
    end
    else
    begin
      //ShowMessage('Abfrage wird abgebrochen');
      MyThread.Terminate;
      // when thread terminates, OnThreadDone will be called (Button1 will then be Enabled)
    end;
  finally
    Screen.Cursor := crDefault;
    Application.ProcessMessages;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  SQLFiles: TStringList;
begin
  SQLFiles := TStringList.Create;
  try
    FindAllFiles(SQLFiles, '.', '*.sql', false); //find e.g. all SQL sourcefiles
    ComboBox1.Items.Assign(SQLFiles);
    ComboBox1.ItemIndex := 0;
  finally
    SQLFiles.Free;
  end;
end;

procedure TForm1.OnThreadDone(Sender: TObject);
begin
  if (MyThread.terminated) then
  begin
    Form1.Caption   := 'TMyThread terminated. Runtime: ' + FormatDateTime('HH:NN:SS', Now - MyThread.fStartTime) + '. Processed rows: ' + IntToStr(MyThread.fRowCnt)
  end
  else
  begin
    Form1.Caption   := 'TMyThread completed. Runtime: ' + FormatDateTime('HH:NN:SS', Now - MyThread.fStartTime) + '. Processed rows: ' + IntToStr(MyThread.fRowCnt);
  end;
  Button1.Caption := 'Start';
  Button1.Enabled := true;

  //ShowMessage('OnThreadDone');
  Application.ProcessMessages;
end;


{ TMyThread }
procedure TMyThread.ButtonStopEnable;
// this method is executed by the mainthread by Synchronize and can therefore access all GUI elements.
begin
  fForm.Button1.Caption := 'Stop';
  fForm.Button1.Enabled := true;
  Application.ProcessMessages;
end;

procedure TMyThread.FetchData;
// this method is executed by the mainthread by Synchronize and can therefore access all GUI elements.
var
  RowData: String;
  i, j:    Integer;

begin
  RowData := '';
  j := 1;
  while (not MyThread.Query.EOF) and (j <= 50) do
  begin
    for i := 0 to MyThread.Query.FieldCount - 2 do
      RowData := RowData + MyThread.Query.Fields[i].AsString + ' - ';
    RowData := RowData + MyThread.Query.Fields[MyThread.Query.FieldCount - 1].AsString;
    fForm.Caption := RowData;
    MyThread.Query.Next;
    RowData := '';
    Inc(j);
    Inc(fRowCnt);
  end;
  Application.ProcessMessages;
end;

procedure TMyThread.ShowStatus;
// this method is executed by the mainthread by Synchronize and can therefore access all GUI elements.
begin
  Form1.Caption := fStatusText;
  Application.ProcessMessages;
end;

constructor TMyThread.Create(CreateSuspended: boolean);
begin
  inherited Create(CreateSuspended);
  FreeOnTerminate := True;
  fSQLText := TStringList.Create;
end;

procedure TMyThread.Execute;
var
  newStatus:    String;
  DatabaseName: String;
  TimerSec:     Integer;
  HardReset:    boolean;

begin
  fStartTime := Now;
  fRowCnt    := 0;

  // Allocate dynamic objects here
  // MyObject := TObject.Create();
  Connection  := TODBCConnection.Create(nil);
  Query       := TSQLQuery.Create(nil);
  Transaction := TSQLTransaction.Create(nil);

  DatabaseName := 'MyDatabase';

  Connection.DatabaseName := DatabaseName;
  Connection.Transaction  := Transaction;
  Query.DataBase          := Connection;

  Query.PacketRecords  := -1;
  Query.ParseSQL       := True;
  Query.ReadOnly       := True;
  Query.UniDirectional := True;

  Query.Close;
  Query.SQL.Clear;

  Query.SQL.Text := fSQLText.Text;

  fStatusText := 'TMyThread Starting...';
  Synchronize(@ShowStatus);

  MyBlockingThread := TMyBlockingThread.Create(True);
  MyBlockingThread.Query := Query;
  MyBlockingThread.Start;

  TimerSec := 0;
  HardReset := False;
  while (not Terminated) and (not MyBlockingThread.Finished) do
  begin
    newStatus := 'TMyThread Timer: ' + FormatDateTime('HH:NN:SS', Now - fStartTime) + '. First loop';
    if NewStatus <> fStatusText then
    begin
      fStatusText := newStatus;
      Synchronize(@ShowStatus);
      if TimerSec = 3 then
        Synchronize(@ButtonStopEnable);
      Inc(TimerSec);
    end;
    sleep(100); // save computing capacity power / energy
    if Terminated then
    begin
      //MyBlockingThread.Terminate;  // Funktioniert nicht!!! Prozessorlast erhöht sich!!!
      TerminateThread(MyBlockingThread.Handle, 0);
      //CloseHandle(MyBlockingThread.Handle);
      //MyBlockingThread := Nil;
      HardReset := True;
    end;
  end;
  //ShowMessage('Raus aus 1. Schleife');
  Synchronize(@ButtonStopEnable);  // You need this, when the loop is finished before timeout (3 sec)!

  while (not Terminated) and (not Query.EOF) do
  begin
    newStatus := 'TMyThread Timer: ' + FormatDateTime('HH:NN:SS', Now - fStartTime) + '. Second loop';
    if NewStatus <> fStatusText then
    begin
      fStatusText := newStatus;
      Synchronize(@ShowStatus);
    end;
    Synchronize(@FetchData);
  end;

  if not Terminated then
  begin
    //MyBlockingThread.Free;
    MyBlockingThread.Terminate;
  end;

  // Destroy dynamic allocated objects here
  // MyObject.Free();
  fSQLText.Free;
  // I can only free this objects, when loop 1 was not interrupted
  // otherwise the application will stop/is blocking.
  if not HardReset then
  begin
    Query.Free;
    Transaction.Free;
    Connection.Free;
  end;
end;

constructor TMyBlockingThread.Create(CreateSuspended: boolean);
begin
  inherited Create(CreateSuspended);
  FreeOnTerminate := True;
end;

procedure TMyBlockingThread.Execute;
begin
  //while true do ;
  Query.Open;
end;

end.

